//
//  sd.c
//
//  SDcard control module mini
//  Copyright (C) 2009 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Minimum read/write function for Secure Digital memory card.
//		Include FAT16 file system, but very slow.
//		The API is like unix system call.
//
//	Spec
//		Physical I/O	: SPI(SCLK=100kHz~400kHz) and CS pin
//		SD card			: under 2GB, not MMC
//		Comunication	: SPI mode
//		CRC check		: Yes
//		Record format	: FAT12 and FAT16 file system
//						  Root directory and short name(8.3) only
//
//	Setting
//		PSoC
//			Device	: CY8C29466
//			SPI		: SPIM user module(Name=SD_SPI,Intrrupt=TXRegEmpty,InvertMISO=normal,Clock=SCLK*2)
//			CS		: LED user module(Name=SD_CS,Drive=Active low)
//			IDE		: PSoC Designer 5 Service Pack 6(Build 1127)
//
//	Reference
//		SDcard	: SD Specifications, Part 1, Physical Layer, Simplified Specification, Version 2.00, September 25, 2006
//				    <www.sdcard.org/developers/tech/sdcard/pls/Simplified_Physical_Layer_Spec.pdf>
//		FAT		: Microsoft Extensible Firmware Initiative, FAT32 File System Specification, FAT: General Overview of On-Disk Format Version 1.03, December 6, 2000
//				    <www.microsoft.com/whdc/system/platform/firmware/fatgen.mspx>
//
//    Update history
//    ---------- ----- -------------------------------------------
//    2009.12.18 v0.0  First cording
//

#include <string.h>						// c library
#include <ctype.h>						// c library
#include "sd.h"							// Private

#ifdef _M8C
#include <m8c.h>						// part specific constants and macros
#include "PSoCAPI.h"					// PSoC API definitions for all User Modules
#include "delay.h"						// Private

//#include "i2c_lcd.h"					// Enable DEBUG
#define printf(a,...) 					// Disable DEBUG
#define i2c_lcd_position(a,b)			// Disable DEBUG
#define i2c_lcd_prCString(a)			// Disable DEBUG
#define i2c_lcd_prHexByte(a)			// Disable DEBUG
#define i2c_lcd_prHexInt(a)				// Disable DEBUG
#endif // _M8C


//*********************************************************************
//
//	C o m m o n
//
//*********************************************************************
// detail error code
char sd_errno = 0;

// Data buffer (Share all layer)
char	sd_buf1[256];					// buf[512], M8C limit is 256
char	sd_buf2[256];

// Convert byte order
unsigned long sd_boconv(unsigned char *buf, char len) {
	unsigned long a;
	unsigned char *p;
	
	a = 0;
	p = buf + len;
	for(; len > 0; --len) {
		a = a << 8 | *--p;
	}
	return(a);
}

void sd_bounconv(unsigned char *buf, char len, unsigned long a) {
	for(; len > 0; --len) {
		*buf++ = a;
		a = a >> 8;
	}
	return;
}

signed char sd_log2(unsigned short x) {
	char i;
	
	for(i = 15; i >= 0; --i) {
		if (x & 0x8000) {
			break;
		}
		x <<= 1;
	}
	return(i);
}

#ifdef SD_CRC_SUPPORT
unsigned char sd_crc7(char *buf, int len) {
	unsigned short	a;				// 0xAABB  A=XORbit B=NextData
	signed char		j;
	
	a = 0;
	for(len += 1; len > 0; --len) {
		if (len > 1) {
			a |= (unsigned char)*buf++;
		}
		for(j = 8; j > 0; --j) {
			if (a & 0x8000) {
				a ^= 0x8900;		// Polynom(high bit on)
			}
			a <<= 1;
		}
	}
	return(a >> 9);
}

unsigned short sd_crc16(char *buf1, char *buf2, int len) {	// len <= 512
	unsigned long	a;					// 0x0AAAAABB  A=XORbit B=NextData
	int				i;
	int				j;
	signed char		k;
	
	a = 0;
	j = len + 2;
	for(i = 0; i < j; i++) {
		if (i < len) {
			if (i & 0x100) {
				a |= (unsigned char)*buf2++;
			} else {
				a |= (unsigned char)*buf1++;
			}
		}
		for (k = 8; k > 0; --k) {
			a <<= 1;
			if (a & 0x01000000) {
				a ^= 0x01102100;		// Polynom(high bit on)
			}
		}
	}
	return(a >> 8);
}
#endif

void sd_memcpy(char *d, char *s, char n) {
	for(; n > 0; --n) {
		*d++ = *s++;
	}
	return;
}

//*********************************************************************
//
// Layer 1	Physical block I/O		(512byte/IO, SPI interface)
//
//*********************************************************************

// Choice sync mode
//#define SD_SPIWAIT()	sd_L1_send(0xff)	// waiting by dummy transmission
//#define SD_SPIWAIT()	delay_50uTimes(1)	// waiting by fixed time
//#define SD_SPIWAIT()	while(!(SD_SPI_bReadStatus() & SD_SPI_SPIM_SPI_COMPLETE))	// SPIM busy
#define SD_SPIWAIT()						// no wait

// SDcard response code
#define SD_R1_IDLE_STATE		0x01
#define SD_R1_ERASE_RESET		0x02
#define SD_R1_ILLEGAL_CMD		0x04
#define SD_R1_CRC_ERR			0x08
#define SD_R1_ERASE_ERR			0x10
#define SD_R1_ADDR_ERR			0x20
#define SD_R1_PRM_ERR			0x40
#define SD_R1_TIMEOUT			0xff


//---------------------------------------------------------------------
//	PSoC SPIM User module interface
//---------------------------------------------------------------------
#ifdef _M8C
//
// Send 1 byte to SPIM user module
//
void sd_L1_send(unsigned char sdata) {

	SD_SPI_SendTxData(sdata);
	while(!(SD_SPI_bReadStatus() & SD_SPI_SPIM_SPI_COMPLETE));

	return;
}

//
// Recieve 1 byte from SPIM user module
//
unsigned char sd_L1_recv(signed char retry) {
										// Recieve data, 0xff=TimeOut
	unsigned char t;
	unsigned char r;

	if (retry > 0) {
		t = 0xff;
	} else {
		t = 0x00;
		retry = -retry;
	}

	for(; retry > 0; --retry) {
		SD_SPI_SendTxData(0xff);
		while(!(SD_SPI_bReadStatus() & SD_SPI_SPIM_SPI_COMPLETE));
		r = SD_SPI_bReadRxData();
		if (r != t) {
			break;
		}
		delay_50uTimes(2);
	}

	return(r);
}
#endif

//---------------------------------------------------------------------
//	SDcard SPI mode control
//---------------------------------------------------------------------
#ifndef UDEBUG
//
// Send SD command and Recieve response
//
unsigned char sd_L1_cmd_r1(char cmd, unsigned long prm) {
										// Response code,  0:normal
#define SD_CMD1_BUFSZ	6
	char			buf[SD_CMD1_BUFSZ];	// In stack
	char			i;
	char			j;
	unsigned char	r;
	
	buf[0] = (cmd & 0x3f) | 0x40;
	buf[1] = prm >> 24;
	buf[2] = prm >> 16;
	buf[3] = prm >> 8;
	buf[4] = prm;
#ifdef SD_CRC_SUPPORT
	buf[5] = (sd_crc7(buf, SD_CMD1_BUFSZ - 1) << 1) | 0x01;
#else
	buf[5] = 0x95;						// CMD0(GO_IDLE_STATE)CRC
#endif

	// Try loop
	for(j = SD_RETRY_CMD; j > 0; --j) {

		// Command
		sd_L1_send(0xff);				// Stable (Kingston accept)
		for(i = 0; i < SD_CMD1_BUFSZ; i++) {
			sd_L1_send(buf[i]);
		}
	
		// Response
		r = sd_L1_recv(SD_WAIT_RESPONSE);
		if ((r & ~SD_R1_IDLE_STATE) == 0x00) {
			break;
		}
	}

	return(r);
}


//
// Open: Initiate SDcard device
//
unsigned char sd_L1_open(void){			// Response code, 0:normal
	int				i;
	unsigned char	r;
	
	// Attach User module
	SD_CS_Start();						// CS contol pin/port
    SD_CS_Off();
    SD_SPI_Start(SD_SPI_SPIM_MODE_0 | SD_SPI_SPIM_MSB_FIRST);
    SD_SPI_EnableInt();
	
	// Power on, wait init sequence
	for(i = SD_WAIT_POWERUP; i > 0; --i) {
		sd_L1_send(0xff);
	}

	// Software reset
	SD_SPIWAIT();
	SD_CS_On();
	SD_SPIWAIT();
	r = sd_L1_cmd_r1(0, 0);				// GO_IDLE_STATE
	SD_SPIWAIT();
	sd_L1_send(0xff);					// stable (go to standby)
	SD_CS_Off();

	// Wait idle process
	for(i = SD_WAIT_INIT_COUNT; i > 0; --i) {

		SD_SPIWAIT();
		SD_CS_On();
		
		SD_SPIWAIT();
		r = sd_L1_cmd_r1(55, 0);		// APP_CMD
//		if(r & ~SD_R1_IDLE_STATE) {
//			break;
//		}
		SD_SPIWAIT();
		r = sd_L1_cmd_r1(41, 0);		// SD_SEND_OP_COND

		SD_SPIWAIT();
		sd_L1_send(0xff);				// stable (go to standby)
		SD_CS_Off();
		
		if (!(r & SD_R1_IDLE_STATE)) {
			break;
		}
		
		delay_1mTimes(SD_WAIT_INIT_INTERVAL);
	}

#ifdef SD_CRC_SUPPORT
	if (r == 0x00) {
		SD_SPIWAIT();
		SD_CS_On();
		SD_SPIWAIT();
		r = sd_L1_cmd_r1(59, 1);		// CRC_ON_OFF
		SD_SPIWAIT();
		sd_L1_send(0xff);				// stable (go to standby)
		SD_CS_Off();
	}
#endif

	i2c_lcd_position(0,0);	i2c_lcd_prCString("L1open="); i2c_lcd_prHexByte(r);
	return(r);
}

//
// Close SDcard device
//
void sd_L1_close(void) {

	// dettach User module
    SD_SPI_DisableInt();
    SD_SPI_Stop();
	SD_CS_Stop();

	i2c_lcd_position(0,0);	i2c_lcd_prCString("L1close ");
	return;
}

//
// Read 1 block into share buffer
//
unsigned char sd_L1_read(long sect){	// Response code, 0:normal
	unsigned char	r;
	char			j;
	unsigned short	c;

	// Try loop
	for(j = SD_RETRY_READ; j > 0; --j) {

		// Transaction start
		SD_SPIWAIT();
		SD_CS_On();
	
		// Dummy loop
		for(;;) {
			char	*p;
			int		i;
		
			// Read cmd
			SD_SPIWAIT();
			r = sd_L1_cmd_r1(17, sect << 9);	// READ_SINGLE_BLOCK
			if (r != 0x00) {
				break;
			}
		
			// Wait start data block talken
			SD_SPIWAIT();
			r = sd_L1_recv(SD_WAIT_READ);
			if (r != 0xfe) {
				break;
			}

			// Recieve, data & CRC
			p = sd_buf1;
			for(i = sizeof(sd_buf1)+sizeof(sd_buf2)+2; i > 0; --i) {
				r = sd_L1_recv(1);		// no retry
				
				if (i == (sizeof(sd_buf1)+2)) {
					p = sd_buf2;
				}
				if (i > 2) {
					*p++ = r;
				} else if (i == 2) {
					c = r << 8;
				} else {
					c |= r;
				}
			}
			
			r = 0x00;
			break;
		}
		
		// Transaction end
		SD_SPIWAIT();
		sd_L1_send(0xff);				// stable (go to standby)
		SD_CS_Off();

		// Normal
		if (r == 0x00) {
#ifdef SD_CRC_SUPPORT
			if (sd_crc16(sd_buf1, sd_buf2, sizeof(sd_buf1)+sizeof(sd_buf2)) != c) {
				r = SD_R1_CRC_ERR;
				continue;
			}
#endif
			break;
		}
	}

	i2c_lcd_position(0,0);	i2c_lcd_prCString("L1read="); i2c_lcd_prHexByte(r);
	return(r);
}

//
// Write 1 block from share buffer
//
unsigned char sd_L1_write(long sect){	// Response code, 0:normal
	unsigned char	r;
	char			j;
	unsigned short	c;

	// Calc CRC
#ifdef SD_CRC_SUPPORT
	c = sd_crc16(sd_buf1, sd_buf2, sizeof(sd_buf1)+sizeof(sd_buf2));
#else
	c = 0;
#endif

	// Try loop
	for(j = SD_RETRY_WRITE; j > 0; --j) {

		// Transaction start
		SD_SPIWAIT();
		SD_CS_On();

		// Dummy loop
		for(;;) {
			char	*p;
			int		i;
		
			// Write cmd
			SD_SPIWAIT();
			r = sd_L1_cmd_r1(24, sect << 9);	// WRITE_BLOCK
			if(r != 0x00) {
				break;
			}
	
			// Start data block talken
			SD_SPIWAIT();
			sd_L1_send(0xfe);
		
			// Send data & CRC
			p = sd_buf1;
			for(i = sizeof(sd_buf1)+sizeof(sd_buf2)+2; i > 0; --i) {
				unsigned char d;
			
				if (i == (sizeof(sd_buf1)+2)) {
					p = sd_buf2;
				}

				d = 0;
				if (i > 2) {
					d = *p++;
				} else if (i == 2) {
					d = c >> 8;
				} else {
					d = c & 0xff;
				}
				sd_L1_send(d);
			}

			// wait response
			r = sd_L1_recv(SD_WAIT_RESPONSE);
			if ((r & 0x1f) != 0x05) {
				break;
			}

			// Busy wait
			r = sd_L1_recv(-SD_WAIT_WRITE);	// wait 0x00 to 0xff
			if (r != 0x00) {
				r = 0;
				break;
			}
		}
		
		// Transaction end
		SD_SPIWAIT();
		sd_L1_send(0xff);				// stable (go to standby)
		SD_CS_Off();

		// Normal
		if (r == 0x00) {
			break;
		}
	}

	i2c_lcd_position(0,0);	i2c_lcd_prCString("L1close="); i2c_lcd_prHexByte(r);
	return(r);
}
#endif // _M8C


#ifdef UDEBUG
// ------------------
//  DEBUG STUB START
// ------------------
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#define i2c_lcd_position(a,b)		// Disable
#define i2c_lcd_prCString(a)		// Disable DEBUG
#define i2c_lcd_prHexByte(a)		// Disable DEBUG
#define i2c_lcd_prHexInt(a)		// Disable DEBUG
int fd1;
unsigned char sd_L1_open(void){
	unsigned char rc;
	
	rc = 0xff;
	for(;;) {
		if((fd1 = open("dump.img", O_RDWR)) < 0) break;
		rc  = 0;
		break;
	}
	printf("L1open()=%d\n",rc);
	return(rc);
}
void sd_L1_close(void) {
	close(fd1);
	printf("L1close()\n");
	return;
}
void debug_dump(void){
	int i;
	
	for(i = 0; i < sizeof(sd_buf1); i++) {
		printf("%02x", (unsigned char)sd_buf1[i]);
		if ((i & 0xf) == 0xf) printf(" ");
		if ((i & 0x1f) == 0x1f) printf("\n");
	}
	for(i = 0; i < sizeof(sd_buf2); i++) {
		printf("%02x", (unsigned char)sd_buf2[i]);
		if ((i & 0xf) == 0xf) printf(" ");
		if ((i & 0x1f) == 0x1f) printf("\n");
	}
}
unsigned char sd_L1_read(long sect){
	unsigned char rc;
	
	rc = 0xff;
	for(;;) {
		if (lseek(fd1, sect*512, SEEK_SET) < 0) break;
		if (read(fd1, sd_buf1, sizeof(sd_buf1)) < 0) break;
		if (lseek(fd1, sect*512+sizeof(sd_buf1), SEEK_SET) < 0) break;
		if (read(fd1, sd_buf2, sizeof(sd_buf2)) < 0) break;
		rc  = 0;
		break;
	}
	printf("L1read(sect=%ld)=%d\n",sect,rc);
	debug_dump();
	return(rc);
}
unsigned char sd_L1_write(long sect){
	unsigned char rc;
	
	rc = 0xff;
	for(;;) {
		if (lseek(fd1, sect*512, SEEK_SET) < 0) break;
		if (write(fd1, sd_buf1, sizeof(sd_buf1)) < 0) break;
		if (lseek(fd1, sect*512+sizeof(sd_buf1), SEEK_SET) < 0) break;
		if (write(fd1, sd_buf2, sizeof(sd_buf2)) < 0) break;
		rc  = 0;
		break;
	}
	printf("L1write(sect=%ld)=%d\n",sect,rc);
	debug_dump();
	return(rc);
}
// ------------------
//  DEBUG STUB END
// ------------------
#endif


//*********************************************************************
//
// Layer 2	Partition manage and data caching
//
//*********************************************************************
#define SD_L2_PASSCODE	0x55
struct sd_L2_part	sd_L2_part[SD_L2_MAXPD] = SD_L2_INIPD;

//
//	Open
//
struct sd_L2_part *sd_L2_open(char partno) {	// discriptor  0:error(detail is sd_errno)
#define SD_L2_TBLSZ 4
	struct parttbl {
		unsigned char	flag;
		unsigned char	start_chs[3];
		unsigned char	type;
		unsigned char	end_chs[3];
		unsigned char	start_lba[4];
		unsigned char	sect[4];
	};
	struct mbr2 {
		unsigned char	bootcode[190];	// +256=446
		struct parttbl	parttbl[SD_L2_TBLSZ];
		unsigned char	bootsig[2];
	};
	struct parttbl	*pt;
	struct mbr2		*m2;
	
	signed char		rc;
	unsigned char	r;
	struct sd_L2_part *pp;
	char			pd;

	rc = -1;
	
	// Dummy loop
	for(;;){

		// Allocate table
		pp = &sd_L2_part[0];
		for(pd = 0; pd < SD_L2_MAXPD; pd++) {
			if (pp->no < 0) {
				break;
			}
			pp++;
		}
		if (pd >= SD_L2_MAXPD) {
			sd_errno = SD_EMFILE;
			break;
		}
		
		// Device open
		if ((r = sd_L1_open()) != 0) {
			sd_errno = SD_EIO;
			if (r == 0xff) {			// Not insert
				sd_errno = SD_EPERM;
			}
			break;
		}

		// Read boot sector
		if (sd_L1_read(0) != 0) {		// Data->L1_buf[]
			sd_errno = SD_EIO;
			break;
		}

		// MBR check
		m2 = (struct mbr2 *)sd_buf2;
		if ((m2->bootsig[0] != 0x55) || (m2->bootsig[1] != 0xAA)) {
			sd_errno = SD_ENODEV;
			break;
		}

		// No entry?
		pt = (struct parttbl *)m2->parttbl;
		if (pt->type == 0) {
			sd_errno = SD_ENOENT;
			break;
		}

		// Set
		pp->parttype  = pt->type;
		pp->startsect = sd_boconv(pt->start_lba, sizeof(pt->start_lba));
		pp->totalsect = sd_boconv(pt->sect, sizeof(pt->sect));
		pp->bufsect   = -1;
		pp->delayw    = 0;
#ifdef SD_L2_ARGCHECK
		pp->sig       = SD_L2_PASSCODE;
#endif
		
		
		printf("L2 pp->parttype=%d\n", pp->parttype);
		printf("L2 pp->startsect=%ld\n", pp->startsect);
		printf("L2 pp->totalsect=%ld\n", pp->totalsect);

		// Commit
		pp->no = pd;
		rc = 0;
		break;
	}
	
	if (rc != 0) {
		pp = 0;
	}

	printf("L2open()=0x%08lx\n", (long)pp);
	i2c_lcd_position(0,0);	i2c_lcd_prCString("L2open="); i2c_lcd_prHexInt((int)pp);
	return(pp);
}

//
//	Close
//
signed char sd_L2_close(struct sd_L2_part *pp) {	// 0:normal -1:error
	signed char rc;
	
	rc = -1;
	
	// dummy loop
	for(;;) {
		
#ifdef SD_L2_ARGCHECK
		if ((pp->no < 0) || (pp->no >= SD_L2_MAXPD) || (pp->sig != SD_L2_PASSCODE)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif

		// Delayed write
		if (pp->delayw) {
			if (sd_L1_write(pp->bufsect + pp->startsect) != 0) {
				sd_errno = SD_EIO;
				break;
			}
		}
		
		// Device close
		sd_L1_close();
		pp->no = -1;
		rc = 0;
		break;
	}
	
	printf("L2close(pp=0x%08lx)=%d\n", (long)pp, rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L2close=");i2c_lcd_prHexInt((int)pp);i2c_lcd_prHexByte(rc);
	return(rc);
}

//
//	Read
//
signed char sd_L2_read(struct sd_L2_part *pp, long sect) {	// 0:normal -1:error(detail is sd_errno)
	signed char rc;
	
	rc = -1;
	
	// dummy loop
	for(;;){
		
#ifdef SD_L2_ARGCHECK
		if ((pp->no < 0) || (pp->no >= SD_L2_MAXPD) || (pp->sig != SD_L2_PASSCODE)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif
		
		// Read Cache
		if (pp->bufsect == sect) {
			rc = 0;
			break;
		}
		
		// Delayed write
		if (pp->delayw) {
			if (sd_L1_write(pp->bufsect + pp->startsect) != 0) {
				sd_errno = SD_EIO;
				break;
			}
			pp->delayw = 0;
		}

#ifdef SD_L2_ARGCHECK
		// Check length
		if (sect >= pp->totalsect) {
			sd_errno = SD_EFBIG;
			break;
		}
#endif

		// Read sector
		if (sd_L1_read(sect + pp->startsect) != 0) {
			sd_errno = SD_EIO;
			break;
		}

		pp->bufsect = sect;
		rc = 0;
		break;
	}

	printf("L2read(pp=0x%08lx,sect=%ld)=%d\n", (long)pp, sect, rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L2read=");i2c_lcd_prHexInt((int)pp);i2c_lcd_prHexInt(sect);i2c_lcd_prHexByte(rc);
	return(rc);
}

//
//	Write 512bytes
//
signed char sd_L2_write(struct sd_L2_part *pp, long sect) {	// 0:normal -1:error(detail is sd_errno)
	signed char rc;
	
	rc = -1;
	
	// dummy loop
	for(;;){
		
#ifdef SD_L2_ARGCHECK
		if ((pp->no < 0) || (pp->no >= SD_L2_MAXPD) || (pp->sig != SD_L2_PASSCODE)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif
		
		// Write Cache
		if (pp->bufsect == sect) {
			pp->delayw = 1;
			rc = 0;
			break;
		}
		
#ifdef SD_L2_ARGCHECK
		// Check length
		if (sect >= pp->totalsect) {
			sd_errno = SD_EFBIG;
			break;
		}
#endif
		
		// Delayed write
		if (pp->delayw) {
			if (sd_L1_write(pp->bufsect + pp->startsect) != 0) {
				sd_errno = SD_EIO;
				break;
			}
			pp->delayw = 0;
		}

		// Marking
		pp->bufsect = sect;
		pp->delayw = 1;

		rc = 0;
		break;
	}

	printf("L2write(pp=0x%08lx,sect=%ld)=%d\n", (long)pp, sect, rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L2write=");i2c_lcd_prHexInt((int)pp);i2c_lcd_prHexInt(sect);i2c_lcd_prHexByte(rc);
	return(rc);
}

//*********************************************************************
//
// Layer 3	FAT file system		Support 12/16bit
//
//*********************************************************************

// External date/time interface
unsigned char	sd_L3_NowTimeTenth = 0;
unsigned char	sd_L3_NowTime[2] = {0, 0};				// 00:00:00
unsigned char	sd_L3_NowDate[2] = {0x81, 0x3b};		// 2009.12.01

//---------------------------------------------------------------------
//	Volume information table
//---------------------------------------------------------------------
// Enable partition ID
#define SD_L3_PARTID_1			0x01		// FAT12
#define SD_L3_PARTID_2			0x04		// FAT16 <32KB
#define SD_L3_PARTID_3			0x06		// FAT16 >32KB
#define SD_L3_PARTID_4			0x0b		// FAT32
#define SD_L3_PARTID_5			0x00		// No partition disk
#define SD_L3_PARTID_6			0x0e		// FAT16 LBA
// Fixed value
#define SD_L3_FATEntSz			2			// 2^1
#define SD_L3_FATEntSzPow		1			// 2^SD_L3_FATEntSzPow == SD_L3_FATEntSz
#define SD_L3_FATEntSzMask		0x01		//
#define SD_L3_RootEntSz			32			// 2^5
#define SD_L3_RootEntSzPow		5			// 2^SD_L3_RootEntSzPow == SD_L3_RootEntSz
#define SD_L3_RootEntSzMask		0x1f		//
#define SD_L3_BytePerSec		512			// 2^9
#define SD_L3_BytePerSecPow		9			// 2^SD_L3_BytePerSecPow == SD_L3_BytePerSec
#define SD_L3_BytePerSecMask	0x01ff		//
#define SD_L3_FAT12				12
#define SD_L3_FAT16				16
#define SD_L3_FAT32				32

struct sd_L3_vol	sd_L3_vol[SD_L3_MAXVD] = SD_L3_INIVD;

//---------------------------------------------------------------------
//	File access control table
//---------------------------------------------------------------------
struct sd_L3_fd		sd_L3_fdtbl[SD_L3_MAXFD] = SD_L3_INIFD;
struct sd_L3_fd		*sd_L3_tempfp;			// define macro temp

//---------------------------------------------------------------------
//	Cluster manage (FAT Region)
//---------------------------------------------------------------------
#define SD_L3_FREECL		0x0000		// Free cluster
#define SD_L3_SPECIALCL		0xfff0		// Functionary
#define SD_L3_BADCL			0xfff7		// Bad cluster
#define SD_L3_EOC			0xffff		// End of cluster
#define SD_L3_READONLYCL	0xfffe		// Pseudo: No update option
#define SD_L3_ROOTDIRCL		0xfffd		// Pseudo: Directory region access

//
// read and update FAT entry
//
unsigned short sd_L3_fatentryio(struct sd_L3_vol *vp, unsigned short cl, unsigned short updcl) {	// read data(SD_L3_BADCL==error)
	unsigned short	d;					// readed data
	
	printf("L3fatentryio(vp=0x%08lx,cl=0x%04x,updcl=0x%04x) start\n", (long)vp, cl, updcl);

	d = SD_L3_BADCL;
	for(;;){
		int		r;						// relative sector location
		int		i;						// buffer index
		long	s;						// absolute sector location
		char	*p;						// buffer pointer

		if (cl >= SD_L3_SPECIALCL) {
			d = cl;
			break;
		}

		if (vp->FATType == SD_L3_FAT12) {
			int	w;

			w = cl + (cl >> 1);			// *1.5
//			r = w / SD_L3_BytePerSec;
//			i = w % SD_L3_BytePerSec;
			r = w >> SD_L3_BytePerSecPow;		// self optimize
			i = w & SD_L3_BytePerSecMask;		// self optimize
			if (i == (SD_L3_BytePerSec - 1)) {	// sector boundery
				i = 0;
				r++;
			}
		} else {						// FAT16
//			r = (cl * SD_L3_FATEntSz) / SD_L3_BytePerSec;
//			i = (cl * SD_L3_FATEntSz) % SD_L3_BytePerSec;
			r = cl >> (SD_L3_BytePerSecPow - SD_L3_FATEntSzPow);		// self optimize
			i = (cl << SD_L3_FATEntSzPow) & SD_L3_BytePerSecMask;		// self optimize
		}

		s = vp->FirstFATSector + r;
		if (sd_L2_read(vp->pp, s) != 0) {
			break;
		}
		if (i & 0x100) {
			p = &sd_buf2[i & 0xff];
		} else {
			p = &sd_buf1[i];
		}
		d = sd_boconv((unsigned char *)p, SD_L3_FATEntSz);
		printf("DEBUG1 d=0x%04x cl=0x%04x\n", d, cl);
		if (vp->FATType == SD_L3_FAT12) {
			if (cl & 1) {
				d >>= 4;
			} else {
				d &= 0x0fff;
			}
			if (d >= (SD_L3_BADCL & 0x0fff)) {
				d |= 0xf000;
			}
			printf("DEBUG2 d=0x%04x cl=0x%04x\n", d, cl);
		}

		// Update write
		if (updcl != SD_L3_READONLYCL) {
			char n;

			if (vp->FATType == SD_L3_FAT12) {
				if (cl & 1) {
					updcl = (updcl << 4) | (*p & 0x0f);
				} else {
					updcl = (updcl & 0x0fff) | ((*(p + 1) & 0xf0) << 8);
				}
			}
			sd_bounconv((unsigned char *)p, SD_L3_FATEntSz, updcl);

			// all copy
			for(n = vp->NumFATs; n > 0; --n) {
				if (sd_L2_write(vp->pp, s) != 0) {
					d = SD_L3_BADCL;
					break;
				}
				s += vp->FATSz;
			}
		}

		break;
	}
	
	printf("L3fatentryio()=0x%04x\n", d);
	return(d);
}

//
// allocate new FAT entry
//
unsigned short sd_L3_alloccl(struct sd_L3_vol *vp, unsigned short firstcl) {
										// new cluster number(SD_L3_BADCL==error)
	unsigned short cl;

	printf("L3alloccl(vp=0x%08lx,firstcl=0x%04x) start\n", (long)vp, firstcl);

	// dummy loop
	for(;;) {
		char found;

		// Find free cluster (sequential search)
		found = 0;
		for(cl = 0; cl < vp->FATEntCnt; cl++) {
			if (sd_L3_fatentryio(vp, cl, SD_L3_READONLYCL) == SD_L3_FREECL) {
				found = 1;
				break;
			}
		}
		if (!found) {
			sd_errno = SD_ENOSPC;
			cl = SD_L3_BADCL;
			break;
		}

		// Allocated mark
		if (sd_L3_fatentryio(vp, cl, SD_L3_EOC) == SD_L3_BADCL) {
			cl = SD_L3_BADCL;
			break;
		}

		// Link chain
		if (firstcl > 1) {	// not root
			unsigned short lastcl;
			unsigned short c;

			for(c = firstcl; (c >= 2) && (c < SD_L3_SPECIALCL); c = sd_L3_fatentryio(vp, c, SD_L3_READONLYCL)) {
				lastcl = c;
			}

			if (sd_L3_fatentryio(vp, lastcl, cl) == SD_L3_BADCL) {
				cl = SD_L3_BADCL;
				break;
			}
		}

		break;
	}

	printf("L3alloccl()=0x%04x\n", cl);
	return(cl);
}

//
// free FAT entry
//
signed char sd_L3_freecl(struct sd_L3_vol *vp, unsigned short firstcl) {
										// 0:normal -1:error
	unsigned short	cl;
	signed char		rc;

	printf("L3freecl(vp=0x%08lx,freecl=0x%04x) start\n", (long)vp, firstcl);
	
	// dummy loop
	for(cl = firstcl; cl < SD_L3_SPECIALCL; ) {
		unsigned short	n;

		if ((n = sd_L3_fatentryio(vp, cl, SD_L3_READONLYCL)) == SD_L3_BADCL) {
			break;
		}

		if (sd_L3_fatentryio(vp, cl, SD_L3_FREECL) == SD_L3_BADCL) {
			break;
		}

		cl = n;
	}
	if (cl != SD_L3_BADCL) {
		rc = 0;
	} else {
		rc = -1;
	}
	
	printf("L3freecl()=%d\n", rc);
	return(rc);
}

//---------------------------------------------------------------------
//	Directory manage (Root Directory Region)
//---------------------------------------------------------------------
struct sd_L3_dirent	sd_L3_tempdirent;		// use reduction stack

//
// formating file name
//
void sd_L3_convdirname(char *out, char *in) {
	char	i;
	char	j;
	static const char ill[] = {
		0x22, 0x2A, 0x2B, 0x2C, 0x2E, 0x2F, 0x3A, 0x3B,
		0x3C, 0x3D, 0x3E, 0x3F, 0x5B, 0x5C, 0x5D, 0x7C,
		0x00 };
	
	for(i = SD_L3_DIRENT_NAMESZ; i > 0; --i) {
		*out = ' ';
		if (*in > ' ') {
			if (*in == '.') {
				if (i > 3) {
					out++;
					continue;
				} else {
					in++;
				}
			}
			*out = toupper(*in++);

			for(j = 0; ill[j] != 0x00; j++) {
				if (*out == ill[j]) {
					*out = '@';
					break;
				}
			}
		}
		out++;
	}
	return;
}


//---------------------------------------------------------------------
//	Data area manage (Data Region)
//---------------------------------------------------------------------
//
// convert seek address to sector number
//
long sd_L3_seek2sect(struct sd_L3_vol *vp, long seek, unsigned short firstcl) {		// sector number  -1:error
	long	s;					// Answer

	printf("L3seek2sect(vp=0x%08lx,seek=%ld,firstcl=0x%04x) start\n", (long)vp, seek, firstcl);
	s = -1;

	if (firstcl == SD_L3_ROOTDIRCL) {
//		s = vp->RootDirSector + (seek / SD_L3_BytePerSec);
		s = vp->RootDirSector + (seek >> SD_L3_BytePerSecPow);
		
	} else {
		unsigned short	n;					// Target cluster position
		unsigned short	cl;					// Current cluster number
		unsigned short	i;

		// cluster position : n
//		n = seek / vp->BytePerClus;
		n = seek >> vp->BytePerClusPow;
		
		// follow cluster chaine : cl
		if ((firstcl == vp->cl_1st) && (n == vp->cl_n)) {	// cache hit
			cl = vp->cl_cl;
			
		} else {
			i = n;
			for(cl = firstcl; (cl >= 2) && (cl < SD_L3_SPECIALCL); cl = sd_L3_fatentryio(vp, cl, SD_L3_READONLYCL)) {
				if (i <= 0) {
					break;
				}
				--i;
			}
		}
		
		// Calculate
		if ((cl >= 2) && (cl < SD_L3_SPECIALCL)) {
//			s = vp->FirstDataSector
//			  + ((cl-2) * vp->SecPerClus)
//		  	  + (seek / SD_L3_BytePerSec % vp->SecPerClus);
			s = vp->FirstDataSector
			  + ((cl-2) << vp->SecPerClusPow)
			  + ((seek >> SD_L3_BytePerSecPow) & vp->SecPerClusMask);
			
			vp->cl_1st = firstcl;		// caching
			vp->cl_n   = n;
			vp->cl_cl  = cl;
		} else {
			if (cl != SD_L3_BADCL) {
				sd_errno = SD_E2BIG;
			}
		}
	}
	

	printf("L3seek2sect()=%ld\n", s);
	return(s);
}

	
//---------------------------------------------------------------------
//	APIs
//---------------------------------------------------------------------

//
// read input
//
int sd_L3_read(struct sd_L3_fd *fp, char *buf, int bufsz) {	// Readed byte length  0>=:EOF or error
	int		rd;
	long 	sek;
	long 	unr;
	

	printf("L3read(fp=0x%08lx,*,bufsz=%d) start\n", (long)fp, bufsz);
	rd  = 0;
	sek = fp->seekpos;
	unr = fp->filesz - sek;
	
	while((bufsz > 0) && (unr > 0)) {
		long	s;
		int		pos;
		int		rem;
		int		i;
		char	*p;
		
#ifdef SD_L3_ARGCHECK
		if ((fp->no < 0) || (fp->no >= SD_L3_MAXFD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif
		
		// Get target sect#
		s = sd_L3_seek2sect(fp->vp, sek, fp->firstcl);
		if (s < 0) {
			rd = -1;
			break;
		}
		
		// Read target sect
		if (sd_L2_read(fp->pp, s) != 0) {
			rd = -1;
			break;
		}
		
		// Transfer to user buffer
		//			pos = sek % SD_L3_BytePerSec;
		pos = sek & SD_L3_BytePerSecMask;
		rem = (SD_L3_BytePerSec - pos) & 0xff;
		if (rem == 0) {
			rem = 256;
		}
		if (rem > unr) {
			rem = unr;
		}
		if (bufsz < rem) {
			rem = bufsz;
		}
		if (pos & 0x100) {
			p = &sd_buf2[pos & 0xff];
		} else {
			p = &sd_buf1[pos];
		}
		sd_memcpy(buf, p, rem);
		
		printf("L3read pos=%d ", pos);
		printf("rem=%d ", rem);
		printf("bufsz=%d ", bufsz);
		printf("sek=%ld ", sek);
		printf("unr=%ld ", unr);
		printf("rd=%d\n", rd);
		
		// seek pos update
		sek   += rem;
		rd    += rem;
		unr   -= rem;
		bufsz -= rem;
		buf   += rem;
	}
	
	if (rd > 0) {
		fp->seekpos += rd;
	}
	
	printf("L3read(*,buf='%-8.8s...',*)=%d\n", (buf-rd), rd);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3read=");i2c_lcd_prHexInt((int)fp);i2c_lcd_prHexInt(rd);
	return(rd);
}

//
// write output
//
int sd_L3_write(struct sd_L3_fd *fp, char *buf, int bufsz) {	// Writed byte length  0>=:error
	int wd;
	
	printf("L3write(fp=0x%08lx,buf='%-8.8s...',bufsz=%d) start\n", (long)fp, buf, bufsz);
	wd = 0;
	
	if (fp->wflg & SD_L3_WFLG_RO) {
		sd_errno = SD_EACCES;
		
	} else {
		long sek;
		
		sek = fp->seekpos;
		while(bufsz > 0) {
			long	s;
			int		pos;
			int		rem;
			char	*p;
			
#ifdef SD_L3_ARGCHECK
			if ((fp->no < 0) || (fp->no >= SD_L3_MAXFD)) {
				sd_errno = SD_EBADF;
				break;
			}
#endif

			// Get target sect#
			while((s = sd_L3_seek2sect(fp->vp, sek, fp->firstcl)) < 0) {
				unsigned short cl;
				
				if (sd_errno != SD_E2BIG) {
					break;
				}
				
				// growth storage
				cl = sd_L3_alloccl(fp->vp, fp->firstcl);
				
				// 1st grow, chaining root
				if (fp->firstcl == 0) {
					fp->firstcl = cl;
					fp->wflg   |= SD_L3_WFLG_WD;
				}
			}
			if (s < 0) {
				break;
			}
			
			// Read target sect
			if (bufsz < SD_L3_BytePerSec) {
				if (sd_L2_read(fp->pp, s) != 0) {
					break;
				}
			}
			
			// Transfer to user buffer
//			pos = sek % SD_L3_BytePerSec;
			pos = sek & SD_L3_BytePerSecMask;
			rem = (SD_L3_BytePerSec - pos) & 0xff;
			if (rem == 0) {
				rem = 256;
			}
			if (bufsz < rem) {
				rem = bufsz;
			}
			if (pos & 0x100) {
				p = &sd_buf2[pos & 0xff];
			} else {
				p = &sd_buf1[pos];
			}
			sd_memcpy(p, buf, rem);
			
			// Rewrite
			if (sd_L2_write(fp->pp, s) != 0) {
				break;
			}
			fp->wflg |= SD_L3_WFLG_WD;
			
			// seek pos update
			sek   += rem;
			wd    += rem;
			bufsz -= rem;
			buf   += rem;
			
		}
	}
	if (wd > 0) {
		fp->seekpos += wd;
		if (fp->filesz < fp->seekpos) {
			fp->filesz = fp->seekpos;
		}
	}
	
	printf("L3write()=%d\n", wd);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3write=");i2c_lcd_prHexInt((int)fp);i2c_lcd_prHexInt(wd);
	return(wd);
}

//
// reposition read/write file offset
//
long sd_L3_lseek(struct sd_L3_fd *fp, long offset, char whence){ // Absolute seek posotion
	long	s;

	printf("L3lseek(fp=0x%08lx,offset=%ld,whence=%d) start\n", (long)fp, offset, whence);
	s = -1;
	
	for(;;){
#ifdef SD_L3_ARGCHECK
		if ((fp->no < 0) || (fp->no >= SD_L3_MAXFD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif

		switch(whence) {
			case SD_SEEK_SET:
				s = offset;
				break;
			case SD_SEEK_CUR:
				s = fp->seekpos + offset;	// enable size < seek
				break;
			case SD_SEEK_END:
				s = fp->filesz;
				break;
			default:
				sd_errno = SD_EINVAL;
				break;
		}
		if (s >= 0) {
			fp->seekpos = s;
		}
		
		break;
	}
	
	printf("L3lseek()=%ld\n", s);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3lseek=");i2c_lcd_prHexInt(s);
	return(s);
}

//
// find directory entry (internal)
//
signed char sd_L3_finddir(struct sd_L3_vol *vp, struct sd_L3_dirent *de, char *dname, char len){	// 1:found 0:not found -1:error
	signed char	rc;
	
	printf("L3finddir(vp=0x%08lx,de=0x%08lx,dname=%-11s,len=%d) start\n", (long)vp, (long)de, dname, len);
	rc = -1;

	if (sd_L3_lseek(vp->dirfp, 0, SD_SEEK_SET) >= 0) {	// Rewind
		while(rc < 0) {
			if (sd_L3_read(vp->dirfp, (char *)de, sizeof(struct sd_L3_dirent)) <= 0) {
				break;
			}
			if (de->Name[0] == 0x00) {
				sd_errno = SD_ENOENT;
				rc = 0;
			}
			if (!strncmp(de->Name, dname, len)) {
				rc = 1;
			}
		}
	}
	
	printf("L3finddir()=%d\n", rc);
	return(rc);
}

//
// open or create a file for reading or writing
//
struct sd_L3_fd *sd_L3_open(struct sd_L3_vol *vp, char *fname, unsigned char oflag, unsigned char mode) {	// File discriptor  -1:error
	signed char		fd;
	struct sd_L3_fd	*fp;
	
	printf("L3open(vp=0x%08lx,file=%s,oflag=0x%02x,mode=0x%02x) start\n", (long)vp, fname, oflag, mode);

	// dummy loop
	for(;;){
		
#ifdef SD_L3_ARGCHECK
		if ((vp->no < 0) || (vp->no >= SD_L3_MAXVD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif

		//
		// attach new fd
		//
		fp = &sd_L3_fdtbl[0];
		for(fd = 0; fd < SD_L3_MAXFD; fd++) {
			if (fp->no < 0) {
				break;
			}
			fp++;
		}
		if (fd >= SD_L3_MAXFD) {
			sd_errno = SD_EMFILE;
			break;
		}
		fp->vp      = vp;
		fp->pp      = vp->pp;
		fp->seekpos = 0;
		fp->wflg    = 0;

		//
		// Root directory file open
		//
		if ((fname[0] == '/') && (fname[1] == '\0')) {
			fp->firstcl = SD_L3_ROOTDIRCL;
			fp->filesz  = vp->RootEntCnt << SD_L3_RootEntSzPow;
			fp->dirloc  = -1;			// Root directory is no entry
			
		//
		// Normal file open
		//
		} else {
			char		m[SD_L3_DIRENT_NAMESZ+1];
			long		dl;				// directory entry location
			signed char	found;
			
			//
			// Format file name
			//
			sd_L3_convdirname(m, fname);
			m[SD_L3_DIRENT_NAMESZ] = '\0';
			if (m[0] == ' ') {
				sd_errno = SD_EINVAL;
				break;
			}
			
			//
			// Directory entry
			//
			found = 0;
			{
				// Search directory entry
				if ((found = sd_L3_finddir(vp, &sd_L3_tempdirent, m, SD_L3_DIRENT_NAMESZ)) < 0) {
					break;
				}
				
				// Create option
				if ((!found) && (oflag & SD_O_CREAT)) {
					char f[1];
					
					// Search free entry
					f[0] = 0xe5;
					if (sd_L3_finddir(vp, &sd_L3_tempdirent, f, 1) < 0) {
						break;
					}
					found = 1;
					
//					memset((char *)&sd_L3_tempdirent, 0, sizeof(dir));
					sd_memcpy(sd_L3_tempdirent.Name, m, sizeof(sd_L3_tempdirent.Name));
					sd_L3_tempdirent.Attr			= mode;
					sd_L3_tempdirent.NTRes			= 0;
					sd_L3_tempdirent.CrtTimeTenth	= sd_L3_NowTimeTenth;
					sd_L3_tempdirent.CrtTime[0]		= sd_L3_NowTime[0];
					sd_L3_tempdirent.CrtTime[1]		= sd_L3_NowTime[1];
					sd_L3_tempdirent.CrtDate[0]		= sd_L3_NowDate[0];
					sd_L3_tempdirent.CrtDate[1]		= sd_L3_NowDate[1];
					sd_L3_tempdirent.LstAccDate[0]	= sd_L3_NowDate[0];
					sd_L3_tempdirent.LstAccDate[1]	= sd_L3_NowDate[1];
					sd_L3_tempdirent.WrtTime[0]		= sd_L3_NowTime[0];
					sd_L3_tempdirent.WrtTime[1]		= sd_L3_NowTime[1];
					sd_L3_tempdirent.WrtDate[0]		= sd_L3_NowDate[0];
					sd_L3_tempdirent.WrtDate[1]		= sd_L3_NowDate[1];
					sd_L3_tempdirent.FstClusLO[0]	= 0;
					sd_L3_tempdirent.FstClusLO[1]	= 0;
					sd_L3_tempdirent.FileSize[0]	= 0;
					sd_L3_tempdirent.FileSize[1]	= 0;
					sd_L3_tempdirent.FileSize[2]	= 0;
					sd_L3_tempdirent.FileSize[3]	= 0;

					// Rewrite
					if (sd_L3_lseek(vp->dirfp, -(long)sizeof(sd_L3_tempdirent), SD_SEEK_CUR) < 0) {
						break;
					}
					if (sd_L3_write(vp->dirfp, (char *)&sd_L3_tempdirent, sizeof(sd_L3_tempdirent)) < 0) {
						break;
					}
				}
				dl = sd_L3_lseek(vp->dirfp, 0, SD_SEEK_CUR) - sizeof(sd_L3_tempdirent);
				
				// Exists
				if ((found) && (oflag & SD_O_EXCL)) {
					sd_errno = SD_EEXIST;
					break;
				}
				if (!found) {
					sd_errno = SD_ENOENT;
					break;
				}
			}
			
			//
			// Init fd table
			//
			
			// setup fd table
			fp->firstcl = sd_boconv(sd_L3_tempdirent.FstClusLO, sizeof(sd_L3_tempdirent.FstClusLO));
			fp->filesz  = sd_boconv(sd_L3_tempdirent.FileSize, sizeof(sd_L3_tempdirent.FileSize));
			fp->dirloc  = dl;
			
			// check file attribute
			if ((sd_L3_tempdirent.Attr & 0x01) || (oflag == SD_O_RDONLY)) {
				fp->wflg |= SD_L3_WFLG_RO;
			}
			
			// truncate
			if (oflag & SD_O_TRUNC) {
				fp->filesz = 0;
			}
			
			// append
			if (oflag & SD_O_APPEND) {
				fp->seekpos = fp->filesz;
			}
		}
		
		
		printf("fd[%d].no=%d\n", fd, fp->no);
		printf("fd[%d].wflg=%d\n", fd, fp->wflg);
		printf("fd[%d].firstcl=%d\n", fd, fp->firstcl);
		printf("fd[%d].filesz=%ld\n", fd, fp->filesz);
		printf("fd[%d].seekpos=%ld\n", fd, fp->seekpos);
		printf("fd[%d].dirloc=%ld\n", fd, fp->dirloc);
		printf("fd[%d].vp=0x%08lx\n", fd, (long)fp->vp);
		printf("fd[%d].pp=0x%08lx\n", fd, (long)fp->pp);
		
		//
		// commit
		//
		fp->no = fd;
		break;
	}

	if (fp->no < 0) {
		fp = 0;
	}
	
	printf("L3open()=0x%08lx errno=%d\n", (long)fp, sd_errno);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3open=");i2c_lcd_prHexInt((int)fp);i2c_lcd_prHexByte(sd_errno);
	return(fp);
}

//
// sync buffer - device
//
signed char sd_L3_sync(struct sd_L3_fd *fp) {	// 0:normal -1:error
	signed char	rc;
	struct sd_L3_vol *vp;

	printf("L3sync(fp=0x%08lx) start\n", (long)fp);
	rc = -1;
	vp = fp->vp;
	
	// dummy loop
	for(;;) {
		
#ifdef SD_L3_ARGCHECK
		if ((fp->no < 0) || (fp->no >= SD_L3_MAXFD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif

		// Directory update
		if ((fp->wflg & SD_L3_WFLG_WD) && (fp->dirloc > 0)) {
			if (sd_L3_lseek(vp->dirfp, fp->dirloc, SD_SEEK_SET) <= 0) {
				break;
			}
			if (sd_L3_read(vp->dirfp, (char *)&sd_L3_tempdirent, sizeof(sd_L3_tempdirent)) <= 0) {
				break;
			}
			sd_bounconv(sd_L3_tempdirent.FstClusLO, sizeof(sd_L3_tempdirent.FstClusLO), fp->firstcl);
			sd_bounconv(sd_L3_tempdirent.FileSize,  sizeof(sd_L3_tempdirent.FileSize),  fp->filesz);
			sd_L3_tempdirent.WrtTime[0] = sd_L3_NowTime[0];
			sd_L3_tempdirent.WrtTime[1] = sd_L3_NowTime[1];
			sd_L3_tempdirent.WrtDate[0] = sd_L3_NowDate[0];
			sd_L3_tempdirent.WrtDate[1] = sd_L3_NowDate[1];
			if (sd_L3_lseek(vp->dirfp, fp->dirloc, SD_SEEK_SET) <= 0) {
				break;
			}
			if (sd_L3_write(vp->dirfp, (char *)&sd_L3_tempdirent, sizeof(sd_L3_tempdirent)) <= 0) {
				break;
			}
			fp->wflg &= ~SD_L3_WFLG_WD;
		}

		// Delayed write trigger
		if (sd_L2_read(vp->pp, 0) != 0) {
			break;
		}
		
		rc = 0;
		break;
	}
	
	printf("L3sync()=%d\n", rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3sync=");i2c_lcd_prHexInt((int)fp);i2c_lcd_prHexByte(rc);
	return(rc);
}

//
// delete a descriptor
//
signed char sd_L3_close(struct sd_L3_fd *fp) {
										// -1:error 0:normal
	signed char	rc;						// return code

	printf("L3close(fp=0x%08lx) start\n", (long)fp);
	rc = -1;
	
	// dummy loop
	for(;;) {

		if (sd_L3_sync(fp) != 0) {
			break;
		}

		// fd free
		fp->no = -1;

		rc = 0;
		break;
	}
	
	printf("L3close()=%d\n", rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3close=");i2c_lcd_prHexInt((int)fp);i2c_lcd_prHexByte(rc);
	return(rc);
}

//
// remove directory entry
//
signed char sd_L3_unlink(struct sd_L3_vol *vp, char *fname){
	signed char rc;
		
	printf("L3unlink(vp=0x%08lx,fname=%s) start\n", (long)vp, fname);
	rc = -1;
	
	// dummy loop
	for(;;) {
		char			m[SD_L3_DIRENT_NAMESZ+1];
		unsigned short	cl;

#ifdef SD_L3_ARGCHECK
		if ((vp->no < 0) || (vp->no >= SD_L3_MAXVD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif

		// Format name
		sd_L3_convdirname(m, fname);
		m[SD_L3_DIRENT_NAMESZ] = '\0';

		// Erase flag on
		if (sd_L3_finddir(vp, &sd_L3_tempdirent, m, SD_L3_DIRENT_NAMESZ) <= 0) {
			break;
		}
		cl = sd_boconv(sd_L3_tempdirent.FstClusLO, sizeof(sd_L3_tempdirent.FstClusLO));
		sd_L3_tempdirent.Name[0] = 0xe5;
		if (sd_L3_lseek(vp->dirfp, -(long)sizeof(sd_L3_tempdirent), SD_SEEK_CUR) < 0) {
			break;
		}
		if (sd_L3_write(vp->dirfp, (char *)&sd_L3_tempdirent, sizeof(sd_L3_tempdirent)) < 0) {
			break;
		}

		// Free cluster chain
		if (sd_L3_freecl(vp, cl) < 0) {
			break;
		}

		rc = 0;
		break;
	}
			
	printf("L3unlink()=%d\n", rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3unlink=");i2c_lcd_prHexInt((int)vp);i2c_lcd_prHexByte(rc);
	return(rc);
}

//
// mount a filesystem
//
struct sd_L3_vol *sd_L3_mount(char partno) {
	signed char			rc;
	signed char			vd;
	struct sd_L3_vol	*vp;
	struct sd_L2_part	*pp;
	
	printf("L3mount(partno=%d) start\n", partno);
	rc = -1;

	// dummy loop
	for(;;) {
		struct bpb1 {
			unsigned char BPB_jmpBoot[3];
			unsigned char BPB_OEMName[8];
			unsigned char BPB_BytePerSec[2];
			unsigned char BPB_SecPerClus;
			unsigned char BPB_RsvdSecCnt[2];
			unsigned char BPB_NumFATs;
			unsigned char BPB_RootEntCnt[2];
			unsigned char BPB_TotSec16[2];
			unsigned char BPB_Media;
			unsigned char BPB_FATSz16[2];
			unsigned char BPB_SecPerTrk[2];
			unsigned char BPB_NumHeads[2];
			unsigned char BPB_HiddSec[4];
			unsigned char BPB_TotSec32[4];
			unsigned char res[220];
		} *bpb1;
		struct bpb2 {
			unsigned char res[254];
			unsigned char sig[2];
		} *bpb2;
		unsigned short	BytePerSec;
//		unsigned char	SecPerClus;
//		unsigned short	RsvdSecCnt;
//		unsigned char	NumFATs;
//		unsigned short	RootEntCnt;
//		unsigned short	TotSec16;
//		unsigned char	Media;
//		unsigned short	FATSz16;
//		unsigned short	SecPerTrk;
//		unsigned short	NumHeads;
//		unsigned long	HiddSec;
//		unsigned long	TotSec32;
		long			TotSec;
		
		unsigned short	RsvdSectors;
		unsigned short	FATSectors;
		unsigned short	RootDirSectors;
		long			DataSectors;
		
		unsigned char	n;
		char			rootdirname[] = "/";
		

		//
		// attach new vd
		//
		vp = &sd_L3_vol[0];
		for(vd = 0; vd < SD_L3_MAXVD; vd++) {
			if (vp->no < 0) {
				break;
			}
			vp++;
		}
		if (vd >= SD_L3_MAXVD) {
			sd_errno = SD_EMFILE;
			break;
		}
		vp->cl_1st = SD_L3_BADCL;		// cluster manage
		vp->cl_n   = 0xffff;			//
		vp->cl_cl  = 0;					//

		// device open
		if ((pp = sd_L2_open(partno)) == 0) {
			break;
		}
		vp->pp = pp;

		// Check target partition
		if ((pp->parttype != SD_L3_PARTID_1) &&
			(pp->parttype != SD_L3_PARTID_2) &&
			(pp->parttype != SD_L3_PARTID_3) &&
			(pp->parttype != SD_L3_PARTID_4) &&
			(pp->parttype != SD_L3_PARTID_5) &&
			(pp->parttype != SD_L3_PARTID_6)) {
			sd_errno = SD_ENOTBLK;
			break;
		}
		
		// get boot sector (reserved region)
		if (sd_L2_read(pp, 0) != 0) {
			break;
		}
		bpb1 = (struct bpb1 *)sd_buf1;
		bpb2 = (struct bpb2 *)sd_buf2;
		if ((bpb2->sig[0] != 0x55) ||
			(bpb2->sig[1] != 0xaa)) {
			sd_errno = SD_ENXIO;
			break;
		}
		
		// set value
		BytePerSec			= sd_boconv(bpb1->BPB_BytePerSec, sizeof(bpb1->BPB_BytePerSec));
		vp->SecPerClus		= bpb1->BPB_SecPerClus;
		RsvdSectors			= sd_boconv(bpb1->BPB_RsvdSecCnt, sizeof(bpb1->BPB_RsvdSecCnt));
		vp->NumFATs			= bpb1->BPB_NumFATs;
		vp->RootEntCnt		= sd_boconv(bpb1->BPB_RootEntCnt, sizeof(bpb1->BPB_RootEntCnt));
		TotSec				= sd_boconv(bpb1->BPB_TotSec16,   sizeof(bpb1->BPB_TotSec16));
		if (TotSec == 0) {
			TotSec			= sd_boconv(bpb1->BPB_TotSec32,   sizeof(bpb1->BPB_TotSec32));
		}
		vp->FATSz			= sd_boconv(bpb1->BPB_FATSz16,    sizeof(bpb1->BPB_FATSz16));
		
		// convert 512byte sector
		n = sd_log2(BytePerSec) - SD_L3_BytePerSecPow;
		vp->FATSz			= vp->FATSz << n;
		vp->SecPerClus		= vp->SecPerClus << n;
		RsvdSectors			= RsvdSectors << n;
		TotSec				= TotSec << n;
		
		// calcurate region size and top location
		FATSectors			= (vp->NumFATs * vp->FATSz);
//		RootDirSectors		= ((vp->RootEntCnt * SD_L3_RootEntSz) + (SD_L3_BytePerSec - 1)) / SD_L3_BytePerSec;
		RootDirSectors		= ((vp->RootEntCnt * SD_L3_RootEntSz) + (SD_L3_BytePerSec - 1)) >> SD_L3_BytePerSecPow;
		DataSectors			= TotSec - (RsvdSectors + FATSectors + RootDirSectors);
		
		vp->FirstFATSector	= RsvdSectors;
		vp->RootDirSector	= RsvdSectors + FATSectors;
		vp->FirstDataSector	= RsvdSectors + FATSectors + RootDirSectors;
		
// calcurate other parameter
		vp->SecPerClusPow	= sd_log2(vp->SecPerClus);
		vp->SecPerClusMask	= ~(0xffff << vp->SecPerClusPow);
//		vp->BytePerClus		= (int)vp->SecPerClus * SD_L3_BytePerSec;
		vp->BytePerClus		= (int)vp->SecPerClus << SD_L3_BytePerSecPow;
		vp->BytePerClusPow	= vp->SecPerClusPow + SD_L3_BytePerSecPow;
//		vp->FATEntCnt		= DataSectors / vp->SecPerClus;		// CountofClusters
		vp->FATEntCnt		= DataSectors >> vp->SecPerClusPow;	// CountofClusters
		if (vp->FATEntCnt < 4085) {
			vp->FATType = SD_L3_FAT12;
		} else if(vp->FATEntCnt < 65525) {
			vp->FATType = SD_L3_FAT16;
		} else {
			vp->FATType = SD_L3_FAT32;
		}
		
		printf("vp->NumFATs=%d\n", vp->NumFATs);
		printf("vp->FATSz=%d\n", vp->FATSz);
		printf("vp->RootEntCnt=%d\n", vp->RootEntCnt);
		printf("vp->SecPerClus=%d\n", vp->SecPerClus);
		printf("RsvdSectors=%d\n", RsvdSectors);
		printf("FATSectors=%d\n", FATSectors);
		printf("RootDirSectors=%d\n", RootDirSectors);
		printf("DataSectors=%ld\n", DataSectors);
		printf("vp->FirstFATSector=%ld\n", vp->FirstFATSector);
		printf("vp->RootDirSector=%ld\n", vp->RootDirSector);
		printf("vp->FirstDataSector=%ld\n", vp->FirstDataSector);
		printf("vp->FATEntCnt=%d\n", vp->FATEntCnt);
		printf("vp->SecPerClusPow=%d\n", vp->SecPerClusPow);
		printf("vp->SecPerClusMask=%d\n", vp->SecPerClusMask);
		printf("vp->BytePerClus=%ld\n", vp->BytePerClus);
		printf("vp->BytePerClusPow=%d\n", vp->BytePerClusPow);
		printf("vp->FATType=%d\n", vp->FATType);
		printf("vp->pp=0x%08lx\n", (long)vp->pp);
		
		if ((vp->RootEntCnt == 0) ||	// FAT32 volume?
			(vp->FATSz      == 0) ||
			(vp->FATType    == SD_L3_FAT32) ||
			((vp->FATType == SD_L3_FAT12) && (n != 0))) {
			sd_errno = SD_ENODEV;
			break;
		}

		// Commit
		vp->no = vd;

		// Open root directory file
		if ((vp->dirfp = sd_L3_open(vp, rootdirname, SD_O_RDWR, SD_ATTR_DIRECTORY)) == 0) {
			break;
		}

		rc = 0;
		
		break;
	}
	
	if (rc) {
		if (vp->dirfp != 0) {
			sd_L3_close(vp->dirfp);
		}
		if (pp != 0) {
			sd_L2_close(pp);
		}
		vp->no = -1;
		vp = 0;
	}

	printf("L3mount()=0x%08lx\n", (long)vp);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3mount=");i2c_lcd_prHexInt((int)vp);i2c_lcd_prHexByte(sd_errno);
	return(vp);
}

//
// dismount a filesystem
//
signed char sd_L3_unmount(struct sd_L3_vol *vp) {
	signed char rc;
	
	printf("L3unmount(0x%08lx) start\n", (long)vp);
	rc = -1;
	
	for (;;) {
		
#ifdef SD_L3_ARGCHECK
		if ((vp->no < 0) || (vp->no >= SD_L3_MAXVD)) {
			sd_errno = SD_EBADF;
			break;
		}
#endif
		
		// Close directory file
		if (vp->dirfp != 0) {
			if (sd_L3_close(vp->dirfp) < 0) {
				break;
			}
			vp->dirfp = 0;
		}

		// Close device
		if (sd_L2_close(vp->pp) != 0) {
			break;
		}

		vp->no = -1;
		rc = 0;
		break;
	}
	
	printf("L3unmount()=%d\n", rc);
	i2c_lcd_position(0,0);i2c_lcd_prCString("L3unmount=");i2c_lcd_prHexInt((int)vp);i2c_lcd_prHexByte(rc);
	return(rc);
}

//------------------------------------ End of source code ---------------------------------------

//
// Unit DEBUG driver
//
#ifdef UDEBUG
main(void)
{
	signed char rc;
	int i;
	int j;
	int fd;
	char fname[10] = "test.txt";
	char buf[80];
	SD_DIR *dp;
	struct sd_dirent *de;
	char dname[10] = "/";
	
	printf("Start...\n");
	
	for(j = 0; j < 1; j++) {
		printf("************** %d *****************\n", j);

	// Mount
	if ((rc = sd_mount(0)) < 0) {
		printf("mount err=%d\n", rc, sd_errno);
		return(1);
	}

	// Open
	if ((fd = sd_open(fname, SD_O_CREAT|SD_O_APPEND, 0)) < 0) {
		printf("open err=%d\n", sd_errno);
		return(1);
	}

	// Write
	for(i = 0; i < 100; i++) {
		sprintf(buf, "line number:%3d \n", i);
		if (sd_write(fd, buf, 17) != 17) {
			printf("write err=%d\n", sd_errno);
			return(1);
		}

		if ((i % 60) == 0) {
			rc = sd_sync(fd);
			if (rc < 0) {
				printf("sync err=%d\n", sd_errno);
				return(1);
			}
		}
	}

	// Close
	rc = sd_close(fd);
	if (rc < 0) {
		printf("close err=%d\n", sd_errno);
		return(1);
	}

	// unmount
	if (sd_unmount()) {
		printf("unmount err=%d\n", sd_errno);
		return(1);
	}

	}

	printf("end\n");
	return(0);
}
#endif

#ifdef AAA
	// Dir
	if ((dp = sd_opendir(dname)) == NULL) {
		printf("opendir errno=%d\n", sd_errno);
		return(1);
	}
	sd_rewinddir(dp);
	while((de = sd_readdir(dp)) != NULL) {
		printf("Name:%-11s Attr:%02x\n", de->Name, de->Attr);
		printf("telldir=%ld\n", sd_telldir(dp));
	}
	if (sd_closedir(dp) < 0) {
		printf("closedir errno=%d\n", sd_errno);
		return(1);
	}

	// Rewind
	if (sd_lseek(fd, 0, SD_SEEK_SET) < 0) {
		printf("lseek err=%d\n", sd_errno);
		return(1);
	}

	// Read
	for(i = 0; i < 10; i++) {
		if (sd_read(fd, buf, 16) != 16) {
			printf("read err=%d\n", sd_errno);
			return(1);
		}
		buf[16] = '\0';
		printf("Read %d: %s\n", i, buf);
	}
#endif
