//
//  sd.h
//
//  SDcard control module mini
//  Copyright (C) 2009 Toyohiko Togashi tog001@nifty.com
//
//  This program is free software; you can redistribute it and/or modify it under the terms of the
//  GNU General Public License as published by the Free Software Foundation; either version 3
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
//  without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//  See the GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License along with this program.
//  If not, see <http://www.gnu.org/licenses/>
//
//
//	Description
//		Minimum read/write function for Secure Digital memory card.
//		Include FAT16 file system, but very slow.
//		The API is like unix system call.
//
//	Spec
//		Physical I/O	: SPI(SCLK=100kHz~400kHz) and CS pin
//		SD card			: under 2GB, not MMC
//		Comunication	: SPI mode
//		CRC check		: Yes
//		Record format	: FAT12 and FAT16 file system
//						  Root directory and short name(8.3) only
//
//	Setting
//		PSoC
//			Device	: CY8C29466
//			SPI		: SPIM user module(Name=SD_SPI,Intrrupt=TXRegEmpty,InvertMISO=normal,Clock=SCLK*2)
//			CS		: LED user module(Name=SD_CS,Drive=Active low)
//			IDE		: PSoC Designer 5 Service Pack 6(Build 1127)
//
//	Reference
//		SDcard	: SD Specifications, Part 1, Physical Layer, Simplified Specification, Version 2.00, September 25, 2006
//					<www.sdcard.org/developers/tech/sdcard/pls/Simplified_Physical_Layer_Spec.pdf>
//		FAT		: Microsoft Extensible Firmware Initiative, FAT32 File System Specification, FAT: General Overview of On-Disk Format Version 1.03, December 6, 2000
//					<www.microsoft.com/whdc/system/platform/firmware/fatgen.mspx>
//
//	Update history
//	---------- ----- -------------------------------------------
//	2009.12.18 v0.0  First cording
//

#ifndef SD_H
#define SD_H

//
//	Option
//
#define SD_CRC_SUPPORT					// 
#define SD_L2_ARGCHECK					// Limit check
#define SD_L3_ARGCHECK					// Limit check

//
// error codes
//
extern char sd_errno;
#define SD_EPERM		1				/* Operation not permitted */	// Not insert
#define SD_ENOENT		2				/* No such file or directory */
#define SD_ESRCH		3				/* No such process */
#define SD_EINTR		4				/* Interrupted system call */
#define SD_EIO			5				/* Input/output error */
#define SD_ENXIO		6				/* Device not configured */		// Unformated
#define SD_E2BIG		7				/* Argument list too long */
#define SD_ENOEXEC		8				/* Exec format error */
#define SD_EBADF		9				/* Bad file descriptor */
#define SD_ECHILD		10				/* No child processes */
#define SD_EDEADLK		11				/* Resource deadlock avoided */ /* 11 was EAGAIN */
#define SD_ENOMEM		12				/* Cannot allocate memory */
#define SD_EACCES		13				/* Permission denied */
#define SD_EFAULT		14				/* Bad address */
#define SD_ENOTBLK		15				/* Block device required */
#define SD_EBUSY		16				/* Device / Resource busy */
#define SD_EEXIST		17				/* File exists */
#define SD_EXDEV		18				/* Cross-device link */
#define SD_ENODEV		19				/* Operation not supported by device */	// Unmatch file system
#define SD_ENOTDIR		20				/* Not a directory */
#define SD_EISDIR		21				/* Is a directory */
#define SD_EINVAL		22				/* Invalid argument */
#define SD_ENFILE		23				/* Too many open files in system */
#define SD_EMFILE		24				/* Too many open files */
#define SD_ENOTTY		25				/* Inappropriate ioctl for device */
#define SD_ETXTBSY		26				/* Text file busy */
#define SD_EFBIG		27				/* File too large */
#define SD_ENOSPC		28				/* No space left on device */
#define SD_ESPIPE		29				/* Illegal seek */
#define SD_EROFS		30				/* Read-only file system */
#define SD_EMLINK		31				/* Too many links */
#define SD_EPIPE		32				/* Broken pipe */


//
// Layer 1	Physical I/O 
//
// SDcard-Machine spec condition
#define SD_WAIT_POWERUP 		10		// 0xff send count
#define SD_WAIT_RESPONSE		10		// Byte transmission time count
#define SD_WAIT_INIT_INTERVAL	10		// ACMD41(initialyze complete check) [ms]
#define SD_WAIT_INIT_COUNT		100		// Total wait = SD_WAIT_INIT_INTERVAL * SD_WAIT_INIT_COUNT
#define SD_WAIT_WRITE			100		// 1count = 100[us] + <1byte TX time>
#define SD_WAIT_READ			100		// 1count = 100[us] + <1byte TX time>
#define SD_RETRY_CMD 			3		// Error retry / Command & Response
#define SD_RETRY_READ			3		// Error retry / 512byte data read
#define SD_RETRY_WRITE			3		// Error retry / 512byte data write


//
// Layer 2	Partition manage
//
// Control table
struct sd_L2_part {
	signed char		no;					// Entry# (-1:free)
	char			delayw;				// Write cache flag
	char			sig;				// Fail safe
	
	unsigned char	parttype;			// ID
	long			startsect;			// Partition top
	long			totalsect;			// Partition size
	long			bufsect;			// Memory buffer in sector
};
#define SD_L2_MAXPD 1					// Maximum open
#define SD_L2_INIPD	{{-1}}				// Initial value
extern struct sd_L2_part	sd_L2_part[SD_L2_MAXPD];


//
// Layer 3	FAT file system
//

// Volume information table
struct sd_L3_vol {
	signed char		no;					// Entry# (-1:free)

	// Region size and location (unit=512byte/sector)
	long			FirstFATSector;		//
	long			RootDirSector;		//
	long			FirstDataSector;	//
	
	// FAT region parameter
	unsigned short	FATEntCnt;			// CountofClusters
	unsigned short	FATSz;				// sect
	unsigned char	NumFATs;			//
	char			FATType;			//
	
	// Root directory region parameter
	unsigned short	RootEntCnt;			//
	
	// Data region parameter
	unsigned short	SecPerClus;
	char			SecPerClusPow;		// 2^SecPerClusPow == SecPerClus
	char			SecPerClusMask;		//
	long 			BytePerClus;		// SecPerClus * SD_L3_BytePerSec
	char			BytePerClusPow;		// 2^SecPerClusPow == SecPerClus

	// Other
	unsigned short	cl_1st;				// Cache memory
	unsigned short	cl_n;				//
	unsigned short	cl_cl;				//
	struct sd_L3_fd		*dirfp;			// directory file, not residence
	struct sd_L2_part	*pp;			// Target partition
};
#define SD_L3_MAXVD 1					// Maximum mount
#define SD_L3_INIVD	{{-1}}				// Initial value
extern struct sd_L3_vol	sd_L3_vol[SD_L3_MAXVD];

// File access control table
struct sd_L3_fd {
	signed char		no;					// Entry# (-1:free)
	unsigned char 	wflg;				// Write flag
	unsigned short	firstcl;			// Cluster number top
	long			filesz;				// Byte length
	long			seekpos;			// Byte address
	long			dirloc;				// Directory entry seek
	struct sd_L3_vol	*vp;			// Target volume
	struct sd_L2_part	*pp;			// Target partition
};
#define SD_L3_MAXFD 4					// Maximum open
#define SD_L3_INIFD	{{-1},{-1},{-1},{-1}}	// Initial value
extern struct sd_L3_fd	sd_L3_fdtbl[SD_L3_MAXFD];
#define SD_L3_WFLG_RO	0x01			// Read only request
#define SD_L3_WFLG_WD	0x02			// Writed

// Directory access
#define SD_L3_DIRENT_NAMESZ	11
struct sd_L3_dirent {					// FAT file system's directory structure
	char			Name[SD_L3_DIRENT_NAMESZ];
	unsigned char	Attr;
	unsigned char	NTRes;
	unsigned char	CrtTimeTenth;
	unsigned char	CrtTime[2];
	unsigned char	CrtDate[2];
	unsigned char	LstAccDate[2];
	unsigned char	FstClusHI[2];
	unsigned char	WrtTime[2];
	unsigned char	WrtDate[2];
	unsigned char	FstClusLO[2];
	unsigned char	FileSize[4];
};

// date/time interface
extern unsigned char	sd_L3_NowTimeTenth;	// Need real time update
extern unsigned char	sd_L3_NowTime[2];	// Need real time update
extern unsigned char	sd_L3_NowDate[2];	// Need real time update

// API
extern struct sd_L3_vol	*sd_L3_mount(char partno);
extern signed char		sd_L3_unmount(struct sd_L3_vol *vp);
extern signed char		sd_L3_unlink(struct sd_L3_vol *vp, char *fname);
extern struct sd_L3_fd	*sd_L3_open(struct sd_L3_vol *vp, char *fname, unsigned char oflag, unsigned char attr);
extern signed char		sd_L3_close(struct sd_L3_fd *fp);
extern int				sd_L3_read(struct sd_L3_fd *fp, char *buf, int bufsz);
extern int				sd_L3_write(struct sd_L3_fd *fp, char *buf, int bufsz);
extern long				sd_L3_lseek(struct sd_L3_fd *fp, long offset, char whence);
extern signed char		sd_L3_sync(struct sd_L3_fd *fp);


//---------------------------------------------------------------------
//	Like a "unix system call" APIs
//---------------------------------------------------------------------
#define SD_DIR								struct sd_L3_fd
#define sd_dirent							sd_L3_dirent
extern struct sd_L3_fd *sd_L3_tempfp;
extern struct sd_L3_dirent 					sd_L3_tempdirent;
#define sd_mount(partno)					(sd_L3_mount(partno) ? 0 : -1)
#define sd_open(filename,oflags,attr)		((sd_L3_tempfp = sd_L3_open(&sd_L3_vol[0], (filename), (oflags), (attr))) != 0 ? sd_L3_tempfp->no : -1)
#define sd_read(fd,buf,bufsize)				sd_L3_read(&sd_L3_fdtbl[(fd)], (buf), (bufsize))
#define sd_write(fd,buf,bufsize)			sd_L3_write(&sd_L3_fdtbl[(fd)], (buf), (bufsize))
#define sd_lseek(fd,offset,whence)			sd_L3_lseek(&sd_L3_fdtbl[(fd)], (offset), (whence))
#define sd_sync(fd)							sd_L3_sync(&sd_L3_fdtbl[(fd)])
#define sd_close(fd)						sd_L3_close(&sd_L3_fdtbl[(fd)])
#define sd_unmount()						sd_L3_unmount(&sd_L3_vol[0])
#define sd_unlink(filename)					sd_L3_unlink(&sd_L3_vol[0], (filename))
#define sd_opendir(dirname)					sd_L3_open(&sd_L3_vol[0], (dirname), SD_O_RDONLY, 0)
#define sd_readdir(dirp)					((sd_L3_read((dirp), (char *)&(sd_L3_tempdirent), sizeof(struct sd_L3_dirent)) > 0) && (sd_L3_tempdirent.Name[0] != '\0') ? &sd_L3_tempdirent : 0)
#define sd_readdir_r(dirp, entry, result)	((sd_L3_read((dirp), (char *)(entry), sizeof(struct sd_L3_dirent)) > 0) && ((entry)->Name[0] != '\0') ? (*(result)=(entry), 0) : (*(result)=0, -1))
#define sd_telldir(dirp)					sd_L3_lseek((dirp), 0, SD_SEEK_CUR)
#define sd_seekdir(dirp, loc)				sd_L3_lseek((dirp), (loc), SD_SEEK_SET)
#define sd_rewinddir(dirp)					sd_L3_lseek((dirp), 0, SD_SEEK_SET)
#define sd_closedir(dirp)					sd_L3_close(dirp)
#define sd_dirfd(dirp)						(dirp->no)

// int sd_mount(void);
// int sd_open(char *filename, int oflags, int attr);
// int sd_read(int fd, char *buf, int bufsize);
// int sd_write(int fd, char *buf, int bufsize);
// int sd_lseek(int fd, long offset, int whence);
// int sd_sync(int fd);
// int sd_close(int fd);
// int sd_unmount(void);
// int sd_unlink(char *filename);
// SD_DIR *sd_opendir(char *dirname);
// struct sd_dirent *sd_readdir(SD_DIR *dirp);
// int sd_readdir_r(SD_DIR *dirp, struct sd_dirent *entry, struct sd_dirent **result));
// long sd_telldir(SD_DIR *dirp);
// void sd_seekdir(SD_DIR *dirp, long loc);
// void sd_rewinddir(SD_DIR *dirp);
// int sd_closedir(SD_DIR *dirp);
// int sd_dirfd(SD_DIR *dirp);

// oflags
#define SD_O_RDONLY			0x00		// open for reading only
#define SD_O_WRONLY			0x01		// open for writing only
#define SD_O_RDWR			0x02		// open for reading and writing
//#define SD_O_NONBLOCK		0xXX		// do not block on open or for data to become available
#define SD_O_APPEND			0x04		// append on each write
#define SD_O_CREAT			0x08		// create file if it does not exist
#define SD_O_TRUNC			0x10		// truncate size to 0
#define SD_O_EXCL			0x20		// error if O_CREAT and the file exists
//#define SD_O_SHLOCK		0xXX		// atomically obtain a shared lock
//#define SD_O_EXLOCK		0xXX		// atomically obtain an exclusive lock
//#define SD_O_NOFOLLOW		0xXX		// do not follow symlinks
//#define SD_O_SYMLINK		0xXX		// allow open of symlinks
//#define SD_O_EVTONLY		0xXX		// descriptor requested for event notifications only

// attr
#define SD_ATTR_READ_ONLY	0x01
#define SD_ATTR_HIDDEN		0x02
#define SD_ATTR_SYSTEM		0x04
#define SD_ATTR_VOLUME_ID	0x08
#define SD_ATTR_DIRECTORY	0x10
#define SD_ATTR_ARCHIVE		0x20
#define SD_ATTR_LONG_NAME	ATTR_READ_ONLY | ATTR_HIDDEN | ATTR_SYSTEM | ATTR_VOLUME_ID

// whence
#define SD_SEEK_SET			0			// the offset is set to offset bytes.
#define SD_SEEK_CUR			1			// the offset is set to its current location plus offset bytes.
#define SD_SEEK_END			2			// the offset is set to the size of the file plus offset bytes.

#endif // SD_H
